@@ -37,9 +37,11 @@ package Mail::SpamAssassin::DnsResolver;
 
 use strict;
 use warnings;
-use bytes;
+# use bytes;
 use re 'taint';
 
+require 5.008001;  # needs utf8::is_utf8()
+
 use Mail::SpamAssassin;
 use Mail::SpamAssassin::Logger;
 use Mail::SpamAssassin::Constants qw(:ip);
@@ -51,7 +53,7 @@ use Time::HiRes qw(time);
 
 our @ISA = qw();
 
-use vars qw($io_socket_module_name);
+our $io_socket_module_name;
 BEGIN {
   if (eval { require IO::Socket::IP }) {
     $io_socket_module_name = 'IO::Socket::IP';
@@ -165,7 +167,9 @@ sub load_resolver {
     dbg("dns: eval failed: $eval_stat");
   };
 
-  dbg("dns: using socket module: %s%s", $io_socket_module_name,
+  dbg("dns: using socket module: %s version %s%s",
+      $io_socket_module_name,
+      $io_socket_module_name->VERSION,
       $self->{force_ipv4} ? ', forced IPv4' :
       $self->{force_ipv6} ? ', forced IPv6' : '');
   dbg("dns: is Net::DNS::Resolver available? %s",
@@ -288,7 +292,7 @@ sub pick_random_available_port {
       Mail::SpamAssassin::Conf::set_ports_range(\$ports_bitset, 0, 0, 0);
       $conf->{dns_available_ports_bitset} = $ports_bitset;
     }
-    # prepare auxilliary data structure to speed up further free-port lookups;
+    # prepare auxiliary data structure to speed up further free-port lookups;
     # 256 buckets, each accounting for 256 ports: 8+8 = 16 bit port numbers;
     # each bucket holds a count of available ports in its range
     my @bucket_counts = (0) x 256;
@@ -302,7 +306,7 @@ sub pick_random_available_port {
       if  ($b eq $all_zeroes) { $ind += 256 }
       elsif ($b eq $all_ones) { $ind += 256; $cnt += 256 }
       else {  # count nontrivial cases the slow way
-        foreach (0..255) { if (vec($ports_bitset, $ind++, 1)) { $cnt++ } }
+        vec($ports_bitset, $ind++, 1) && $cnt++  for 0..255;
       }
       $available_portscount += $cnt;
       $bucket_counts[$bucket] = $cnt;
@@ -387,8 +391,6 @@ sub connect_sock {
   } else {  # unrecognized
     # unspecified address, unspecified protocol family
   }
-  dbg("dns: LocalAddr: %s, name server(s): %s",
-      $srcaddr||'', join(', ',@ns_addr_port));
 
   # find a free local random port from a set of declared-to-be-available ports
   my $lport;
@@ -405,6 +407,8 @@ sub connect_sock {
       $errno = 0;
       last;
     }
+    dbg("dns: LocalAddr: [%s]:%d, name server: [%s]:%d, module %s",
+        $srcaddr||'x', $lport,  $ns_addr, $ns_port,  $io_socket_module_name);
     my %args = (
         PeerAddr => $ns_addr,
         PeerPort => $ns_port,
@@ -543,7 +547,7 @@ sub new_dns_packet {
   eval {
 
     if (utf8::is_utf8($domain)) {  # since Perl 5.8.1
-      info("dns: new_dns_packet: domain is utf8 flagged: %s", $domain);
+      dbg("dns: new_dns_packet: domain is utf8 flagged: %s", $domain);
     }
 
     $domain =~ s/\.*\z/./s;
@@ -577,17 +581,19 @@ sub new_dns_packet {
     #    time, $domain, $type, $packet->id);
     1;
   } or do {
-    # this can if a domain name in a query is invalid, or if a timeout signal
+    # get here if a domain name in a query is invalid, or if a timeout signal
     # happened to be trapped by this eval, or if Net::DNS signalled an error
     my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
     # resignal if alarm went off
     die "dns: (1) $eval_stat\n"  if $eval_stat =~ /__alarm__ignore__\(.*\)/s;
-    warn sprintf(
-           "dns: new_dns_packet (domain=%s type=%s class=%s) failed: %s\n",
+    info("dns: new_dns_packet (domain=%s type=%s class=%s) failed: %s",
            $domain, $type, $class, $eval_stat);
   };
 
   if ($packet) {
+    # RD flag needs to be set explicitly since Net::DNS 1.01, Bug 7223	
+    $packet->header->rd(1);
+
   # my $udp_payload_size = $self->{res}->udppacketsize;
     my $udp_payload_size = $self->{conf}->{dns_options}->{edns};
     if ($udp_payload_size && $udp_payload_size > 512) {
@@ -718,6 +724,37 @@ sub bgsend {
 
 ###########################################################################
 
+=item $id = $res->bgread()
+
+Similar to C<Net::DNS::Resolver::bgread>.  Reads a DNS packet from
+a supplied socket, decodes it, and returns a Net::DNS::Packet object
+if successful.  Dies on error.
+
+=cut
+
+sub bgread {
+  my ($self) = @_;
+  my $sock = $self->{sock};
+  my $packetsize = $self->{res}->udppacketsize;
+  $packetsize = 512  if $packetsize < 512;  # just in case
+  my $data = '';
+  my $peeraddr = $sock->recv($data, $packetsize+256);  # with some size margin for troubleshooting
+  defined $peeraddr or die "bgread: recv() failed: $!";
+  my $peerhost = $sock->peerhost;
+  $data ne '' or die "bgread: received empty packet from $peerhost";
+  dbg("dns: bgread: received %d bytes from %s", length($data), $peerhost);
+  my($answerpkt, $decoded_length) = Net::DNS::Packet->new(\$data);
+  $answerpkt or die "bgread: decoding DNS packet failed: $@";
+  $answerpkt->answerfrom($peerhost);
+  if (defined $decoded_length && $decoded_length ne "" && $decoded_length != length($data)) {
+    warn sprintf("bgread: received a %d bytes packet from %s, decoded %d bytes\n",
+                 length($data), $peerhost, $decoded_length);
+  }
+  return $answerpkt;
+}
+
+###########################################################################
+
 =item $nfound = $res->poll_responses()
 
 See if there are any C<bgsend> reply packets ready, and return
@@ -765,13 +802,24 @@ sub poll_responses {
     $timeout = 0;  # next time around collect whatever is available, then exit
     last  if $nfound == 0;
 
-    my $packet = $self->{res}->bgread($self->{sock});
+    my $packet;
+    # Bug 7265, use our own bgread() below
+    # $packet = $self->{res}->bgread($self->{sock});
+    eval {
+      $packet = $self->bgread();  # Bug 7265, use our own bgread()
+    } or do {
+      undef $packet;
+      my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
+      # resignal if alarm went off
+      die $eval_stat  if $eval_stat =~ /__alarm__ignore__\(.*\)/s;
+      info("dns: bad dns reply: %s", $eval_stat);
+    };
 
     if (!$packet) {
-      my $dns_err = $self->{res}->errorstring;
-      # resignal if alarm went off
-      die "dns (3) $dns_err\n"  if $dns_err =~ /__alarm__ignore__\(.*\)/s;
-      info("dns: bad dns reply: $dns_err");
+      # error already reported above
+#     my $dns_err = $self->{res}->errorstring;
+#     die "dns (3) $dns_err\n"  if $dns_err =~ /__alarm__ignore__\(.*\)/s;
+#     info("dns: bad dns reply: $dns_err");
     } else {
       my $header = $packet->header;
       if (!$header) {
@@ -807,8 +855,13 @@ sub poll_responses {
           $cb->($packet, $id, $now);
           $cnt++;
         } else {  # no match, report the problem
-          info("dns: no callback for id %s, ignored; packet: %s",
-               $id,  $packet ? $packet->string : "undef" );
+          if ($rcode eq 'REFUSED' || $id =~ m{^\d+/NO_QUESTION_IN_PACKET\z}) {
+            # the failure was already reported above
+          } else {
+            info("dns: no callback for id $id, ignored, packet on next debug line");
+            # prevent filling normal logs with huge packet dumps
+            dbg("dns: %s", $packet ? $packet->string : "undef");
+          }
           # report a likely matching query for diagnostic purposes
           local $1;
           if ($id =~ m{^(\d+)/}) {
@@ -853,7 +906,8 @@ Emulates C<Net::DNS::Resolver::send()>.
 This subroutine is a simple synchronous leftover from SpamAssassin version
 3.3 and does not participate in packet query caching and callback grouping
 as implemented by AsyncLoop::bgsend_and_start_lookup().  As such it should
-be avoided for mainstream usage.
+be avoided for mainstream usage.  Currently used through Mail::SPF::Server
+by the SPF plugin.
 
 =cut
 
