@@ -46,15 +46,18 @@ use strict;
 use warnings;
 use re 'taint';
 
+BEGIN {
+  eval { require Digest::SHA; import Digest::SHA qw(sha1 sha1_hex); 1 }
+  or do { require Digest::SHA1; import Digest::SHA1 qw(sha1 sha1_hex) }
+}
+
 use Mail::SpamAssassin;
 use Mail::SpamAssassin::Message::Node;
 use Mail::SpamAssassin::Message::Metadata;
 use Mail::SpamAssassin::Constants qw(:sa);
 use Mail::SpamAssassin::Logger;
 
-use vars qw(@ISA);
-
-@ISA = qw(Mail::SpamAssassin::Message::Node);
+our @ISA = qw(Mail::SpamAssassin::Message::Node);
 
 # ---------------------------------------------------------------------------
 
@@ -63,10 +66,10 @@ use vars qw(@ISA);
 Creates a Mail::SpamAssassin::Message object.  Takes a hash reference
 as a parameter.  The used hash key/value pairs are as follows:
 
-C<message> is either undef (which will use STDIN), a scalar of the
-entire message, an array reference of the message with 1 line per array
-element, and either a file glob or IO::File object which holds the entire
-contents of the message.
+C<message> is either undef (which will use STDIN), a scalar - a string
+containing an entire message, a reference to such string, an array reference
+of the message with one line per array element, or either a file glob
+or an IO::File object which holds the entire contents of the message.
 
 Note: The message is expected to generally be in RFC 2822 format, optionally
 including an mbox message separator line (the "From " line) as the first line.
@@ -110,12 +113,34 @@ sub new {
   my $self = $class->SUPER::new({normalize=>$normalize});
 
   $self->{tmpfiles} =           [];
+  $self->{pristine_msg} = 	'';
   $self->{pristine_headers} =	'';
   $self->{pristine_body} =	'';
   $self->{mime_boundary_state} = {};
   $self->{line_ending} =	"\012";
   $self->{master_deadline} = $opts->{'master_deadline'};
   $self->{suppl_attrib} = $opts->{'suppl_attrib'};
+  $self->{body_part_scan_size} = $opts->{'body_part_scan_size'} || 0;
+  $self->{rawbody_part_scan_size} = $opts->{'rawbody_part_scan_size'} || 0;
+
+  if ($self->{suppl_attrib}) {  # caller-provided additional information
+    # pristine_body_length is currently used by an eval test check_body_length
+    # Possible To-Do: Base the length on the @message array later down?
+    if (defined $self->{suppl_attrib}{body_size}) {
+      # Optional info provided by a caller; should reflect the original
+      # message body size if provided, and as such it may differ from the
+      # $self->{pristine_body} size, e.g. when the caller passed a truncated
+      # message to SpamAssassin, or when counting line-endings differently.
+      $self->{pristine_body_length} = $self->{suppl_attrib}{body_size};
+    }
+    if (ref $self->{suppl_attrib}{mimepart_digests}) {
+      # Optional info provided by a caller: an array of digest codes (e.g. SHA1)
+      # of each MIME part. Should reflect the original message if provided.
+      # As such it may differ from digests calculated by get_mimepart_digests(),
+      # e.g. when the caller passed a truncated message to SpamAssassin.
+      $self->{mimepart_digests} = $self->{suppl_attrib}{mimepart_digests};
+    }
+  }
 
   bless($self,$class);
 
@@ -139,14 +164,26 @@ sub new {
       # and is faster than (<$message>) by 10..25 %
       # (a drawback is a short-term double storage of a text in $raw_str)
       #
-      my($inbuf,$nread,$raw_str); $raw_str = '';
-      while ( $nread=sysread($message,$inbuf,16384) ) { $raw_str .= $inbuf }
+      my($nread,$raw_str); $raw_str = '';
+      while ( $nread=sysread($message, $raw_str, 16384, length $raw_str) ) { }
       defined $nread  or die "error reading: $!";
       @message = split(/^/m, $raw_str, -1);
 
-      dbg("message: empty message read")  if $raw_str eq '';
+      if ($raw_str eq '') {
+        dbg("message: empty message read");
+      } elsif (length($raw_str) > 128*1024) {
+        # ditch rarely used large chunks of allocated memory, Bug 6514
+        #   http://www.perlmonks.org/?node_id=803515
+        # about 97% of mail messages are below 128 kB,
+        # about 98% of mail messages are below 256 kB (2010 statistics)
+        # dbg("message: deallocating %.2f MB", length($raw_str)/1024/1024);
+        undef $raw_str;
+      }
     }
   }
+  elsif (ref $message eq 'SCALAR') {
+    @message = split(/^/m, $$message, -1);
+  }
   elsif (ref $message) {
     dbg("message: Input is a reference of unknown type!");
   }
@@ -154,16 +191,36 @@ sub new {
     @message = split(/^/m, $message, -1);
   }
 
-  # Pull off mbox and mbx separators
-  # also deal with null messages
+  # Deal with null message
   if (!@message) {
     # bug 4884:
     # if we get here, it means that the input was null, so fake the message
     # content as a single newline...
     @message = ("\n");
-  } elsif ($message[0] =~ /^From\s+(?!:)/) {
+  }
+
+  # Bug 7648:
+  # Make sure the message is tainted. When linting, @testmsg is not, so this
+  # handles that. Perhaps 3rd party tools could call this with untainted
+  # messages? Tainting the message is important because it prevents certain
+  # exploits later.
+  if (Mail::SpamAssassin::Util::am_running_in_taint_mode() &&
+        grep { !Scalar::Util::tainted($_) } @message) {
+    local($_);
+    # To preserve newlines, no joining and splitting here, process each line
+    # directly as is.
+    foreach (@message) {
+      $_ = Mail::SpamAssassin::Util::taint_var($_);
+    }
+    if (grep { !Scalar::Util::tainted($_) } @message) {
+      die "Mail::SpamAssassin::Message failed to enforce message taintness";
+    }
+  }
+
+  # Pull off mbox and mbx separators
+  if ($message[0] =~ /^From\s+(?!:)/) {
     # careful not to confuse with obsolete syntax which allowed WSP before ':'
-    # mbox formated mailbox
+    # mbox formatted mailbox
     $self->{'mbox_sep'} = shift @message;
   } elsif ($message[0] =~ MBX_SEPARATOR) {
     $_ = shift @message;
@@ -196,12 +253,15 @@ sub new {
 
   # bug 4363
   # Check to see if we should do CRLF instead of just LF
-  # For now, just check the first header and do whatever it does
-  if (@message && $message[0] =~ /\015\012/) {
+  # For now, just check the first and last line and do whatever it does
+  if (@message && ($message[0] =~ /\015\012/ || $message[-1] =~ /\015\012/)) {
     $self->{line_ending} = "\015\012";
     dbg("message: line ending changed to CRLF");
   }
 
+  # Is a CRLF -> LF line endings conversion necessary?
+  my $squash_crlf = $self->{line_ending} eq "\015\012";
+
   # Go through all the header fields of the message
   my $hdr_errors = 0;
   my $header;
@@ -228,6 +288,8 @@ sub new {
 
         # If it's not a valid header (aka: not in the form "foo:bar"), skip it.
         if (defined $value) {
+	  # CRLF -> LF line-endings conversion if necessary
+	  $value =~ s/\015\012/\012/sg  if $squash_crlf;
 	  $key =~ s/[ \t]+\z//;  # strip WSP before colon, obsolete rfc822 syn
 	  # limit the length of the pairs we store
 	  if (length($key) > MAX_HEADER_KEY_LENGTH) {
@@ -256,18 +318,21 @@ sub new {
  	last;
       }
       # should we assume entering a body on encountering invalid header field?
-      elsif ($current !~ /^[\041-\071\073-\176]+[ \t]*:/) {
-	# A field name MUST be composed of printable US-ASCII characters
-	# (i.e., characters that have values between 33 (041) and 126 (176),
-	# inclusive), except colon (072). Obsolete header field syntax
-	# allowed WSP before a colon.
-	if (++$hdr_errors <= 3) {
-	  # just consume but ignore a few invalid header fields
-	} else {  # enough is enough...
-	  $self->{'missing_head_body_separator'} = 1;
-	  unshift(@message, $current);
- 	  last;
- 	}
+      else {
+        # no re "strict";  # since perl 5.21.8: Ranges of ASCII printables...
+        if ($current !~ /^[\041-\071\073-\176]+[ \t]*:/) {
+	  # A field name MUST be composed of printable US-ASCII characters
+	  # (i.e., characters that have values between 33 (041) and 126 (176),
+	  # inclusive), except colon (072). Obsolete header field syntax
+	  # allowed WSP before a colon.
+	  if (++$hdr_errors <= 3) {
+	    # just consume but ignore a few invalid header fields
+	  } else {  # enough is enough...
+	    $self->{'missing_head_body_separator'} = 1;
+	    unshift(@message, $current);
+ 	    last;
+	  }
+	}
       }
 
       # start collecting a new header field
@@ -281,15 +346,23 @@ sub new {
   # will get modified below
   $self->{'pristine_body'} = join('', @message);
 
-  # CRLF -> LF
-  # also merge multiple blank lines into a single one
-  my $start;
+  if (!defined $self->{pristine_body_length}) {
+    $self->{'pristine_body_length'} = length $self->{'pristine_body'};
+  }
+
+  # Store complete message, get_pristine() is used a lot, avoid making copies
+  $self->{'pristine_msg'} = $self->{'pristine_headers'} . $self->{'pristine_body'};
+
   # iterate over lines in reverse order
+  # merge multiple blank lines into a single one
+  my $start;
   for (my $cnt=$#message; $cnt>=0; $cnt--) {
-    $message[$cnt] =~ s/\015\012/\012/;
+    # CRLF -> LF line-endings conversion if necessary
+    $message[$cnt] =~ s/\015\012\z/\012/  if $squash_crlf;
 
     # line is blank
-    if ($message[$cnt] !~ /\S/) {
+    if ($message[$cnt] =~ /^\s*$/) {
+      # /^\s*$/ is about 5% faster then !/\S/, but still expensive here
       if (!defined $start) {
         $start=$cnt;
       }
@@ -300,9 +373,10 @@ sub new {
 
     # if we've got a series of blank lines, get rid of them
     if (defined $start) {
+      my $max_blank_lines = 20;
       my $num = $start-$cnt;
-      if ($num > 10) {
-        splice @message, $cnt+2, $num-1;
+      if ($num > $max_blank_lines) {
+        splice @message, $cnt+2, $num-$max_blank_lines;
       }
       undef $start;
     }
@@ -313,6 +387,24 @@ sub new {
   ($self->{'type'}, $boundary) = Mail::SpamAssassin::Util::parse_content_type($self->header('content-type'));
   dbg("message: main message type: ".$self->{'type'});
 
+#  dbg("message: \$message[0]: \"" . $message[0] . "\"");
+
+  # bug 6845: if main message type is multipart and the message body does not begin with
+  # either a blank line or the boundary (if defined), insert a blank line
+  # to ensure proper parsing - do not consider MIME headers at the beginning of the body
+  # to be part of the message headers.
+  if ($self->{'type'} =~ /^multipart\//i && $#message > 0 && $message[0] =~ /\S/)
+  {
+    if (!defined $boundary || $message[0] !~ /^--\Q$boundary\E/)
+    {
+      dbg("message: Inserting blank line at top of body to ensure correct multipart MIME parsing");
+      unshift(@message, "\012");
+    }
+  }
+
+#  dbg("message: \$message[0]: \"" . $message[0] . "\"");
+#  dbg("message: \$message[1]: \"" . $message[1] . "\"");
+
   # parse queue, simple array of parts to parse:
   # 0: part object, already in the tree
   # 1: boundary used to focus body parsing
@@ -429,7 +521,7 @@ Returns a scalar of the entire pristine message.
 
 sub get_pristine {
   my ($self) = @_;
-  return $self->{pristine_headers} . $self->{pristine_body};
+  return $self->{pristine_msg};
 }
 
 =item get_pristine_body()
@@ -453,7 +545,7 @@ sub extract_message_metadata {
   my ($self, $permsgstatus) = @_;
 
   # do this only once per message, it can be expensive
-  if ($self->{already_extracted_metadata}) { return; }
+  return  if $self->{already_extracted_metadata};
   $self->{already_extracted_metadata} = 1;
 
   $self->{metadata}->extract ($self, $permsgstatus);
@@ -555,11 +647,17 @@ sub finish {
   delete $self->{'mime_boundary_state'};
   delete $self->{'mbox_sep'};
   delete $self->{'normalize'};
+  delete $self->{'body_part_scan_size'};
+  delete $self->{'rawbody_part_scan_size'};
+  delete $self->{'pristine_msg'};
   delete $self->{'pristine_body'};
   delete $self->{'pristine_headers'};
   delete $self->{'line_ending'};
   delete $self->{'missing_head_body_separator'};
 
+  # Remove the queue variable, in case the body has not been parsed
+  delete $self->{'parse_queue'};
+
   my @toclean = ( $self );
 
   # Go ahead and clean up all of the Message::Node parts
@@ -672,25 +770,34 @@ sub parse_body {
     # one doesn't, assume it's malformed and send it to be parsed as a
     # non-multipart section
     #
-    if ( $toparse->[0]->{'type'} =~ /^multipart\//i && defined $toparse->[1] && ($toparse->[3] > 0)) {
+    my ($msg, $boundary, $body, $subparse) = @$toparse;
+
+    if ($msg->{'type'} =~ m{^multipart/}i && defined $boundary && $subparse > 0) {
       $self->_parse_multipart($toparse);
     }
     else {
       # If it's not multipart, go ahead and just deal with it.
       $self->_parse_normal($toparse);
 
-      # bug 5041: exclude message/partial messages, however
-      if ($toparse->[0]->{'type'} =~ /^message\b/i &&
-          $toparse->[0]->{'type'} !~ /^message\/partial$/i &&
-            ($toparse->[3] > 0))
+      # bug 5041: process message/*, but exclude message/partial content types
+      if ($msg->{'type'} =~ m{^message/(?!partial\z)}i && $subparse > 0)
       {
-        # Just decode the part, but we don't care about the result here.
-        $toparse->[0]->decode(0);
+        # Just decode the part, but we don't need the resulting string here.
+        $msg->decode(0);
 
-        # bug 5051, bug 3748: sometimes message/* parts have no content,
-        # and we get stuck waiting for STDIN, which is bad. :(
-        if (defined $toparse->[0]->{'decoded'} &&
-            $toparse->[0]->{'decoded'} ne '')
+        # bug 7125: decode and parse only message/rfc822 or message/global,
+        # but do not treat other message/* content types (like the ones listed
+        # here) as a message consisting of a header and a body, as they are not:
+        #    message/delivery-status, message/global-delivery-status,
+        #    message/feedback-report, message/global-headers,
+        #    message/global-disposition-notification,
+        #    message/disposition-notification, (and message/partial)
+
+        # bug 5051, bug 3748: check $msg->{decoded}: sometimes message/* parts
+        # have no content, and we get stuck waiting for STDIN, which is bad. :(
+
+        if ($msg->{'type'} =~ m{^message/(?:rfc822|global)\z}i &&
+            defined $msg->{'decoded'} && $msg->{'decoded'} ne '')
         {
 	  # Ok, so this part is still semi-recursive, since M::SA::Message
 	  # calls M::SA::Message, but we don't subparse the new message,
@@ -699,14 +806,14 @@ sub parse_body {
 	  # since it's faster.
 	  # 
           my $msg_obj = Mail::SpamAssassin::Message->new({
-    	    message	=>	$toparse->[0]->{'decoded'},
+    	    message	=>	$msg->{'decoded'},
 	    parsenow	=>	0,
 	    normalize	=>	$self->{normalize},
-	    subparse	=>	$toparse->[3]-1,
+	    subparse	=>	$subparse - 1,
 	    });
 
 	  # Add the new message to the current node
-          $toparse->[0]->add_body_part($msg_obj);
+          $msg->add_body_part($msg_obj);
 
 	  # now this is the sneaky bit ... steal the sub-message's parse_queue
 	  # and add it to ours.  then we'll handle the sub-message in our
@@ -717,15 +824,15 @@ sub parse_body {
 	  # Ok, we've subparsed, so go ahead and remove the raw and decoded
 	  # data because we won't need them anymore (the tree under this part
 	  # will have that data)
-	  if (ref $toparse->[0]->{'raw'} eq 'GLOB') {
+	  if (ref $msg->{'raw'} eq 'GLOB') {
 	    # Make sure we close it if it's a temp file -- Bug 5166
-	    close($toparse->[0]->{'raw'})
+	    close($msg->{'raw'})
 	      or die "error closing input file: $!";
 	  }
 
-	  delete $toparse->[0]->{'raw'};
+	  delete $msg->{'raw'};
 	  
-	  delete $toparse->[0]->{'decoded'};
+	  delete $msg->{'decoded'};
         }
       }
     }
@@ -766,13 +873,16 @@ sub _parse_multipart {
     my $line;
     my $tmp_line = @{$body};
     for ($line=0; $line < $tmp_line; $line++) {
+#     dbg("message: multipart line $line: \"" . $body->[$line] . "\"");
       # specifically look for an opening boundary
-      if ($body->[$line] =~ /^--\Q$boundary\E\s*$/) {
+      if (substr($body->[$line],0,2) eq '--'  # triage
+          && $body->[$line] =~ /^--\Q$boundary\E\s*$/) {
 	# Make note that we found the opening boundary
 	$self->{mime_boundary_state}->{$boundary} = 1;
 
 	# if the line after the opening boundary isn't a header, flag it.
 	# we need to make sure that there's actually another line though.
+	# no re "strict";  # since perl 5.21.8: Ranges of ASCII printables...
 	if ($line+1 < $tmp_line && $body->[$line+1] !~ /^[\041-\071\073-\176]+:/) {
 	  $self->{'missing_mime_headers'} = 1;
 	}
@@ -794,14 +904,33 @@ sub _parse_multipart {
   my $in_body = 0;
   my $header;
   my $part_array;
+  my $found_end_boundary;
+  my $partcnt = 0;
 
   my $line_count = @{$body};
   foreach ( @{$body} ) {
     # if we're on the last body line, or we find any boundary marker,
-    # deal with the mime part
-    if ( --$line_count == 0 || (defined $boundary && /^--\Q$boundary\E(?:--)?\s*$/) ) {
+    # deal with the mime part;
+    # a triage before an unlikely-to-match regexp avoids a CPU hotspot
+    $found_end_boundary = defined $boundary && substr($_,0,2) eq '--'
+                          && /^--\Q$boundary\E(?:--)?\s*$/;
+    if ( --$line_count == 0 || $found_end_boundary ) {
       my $line = $_; # remember the last line
 
+      # If at last line and no end boundary found, the line belongs to body
+      # TODO:
+      #  Is $self->{mime_boundary_state}->{$boundary}-- needed here?
+      #  Could "missing end boundary" be a useful rule? Mark it somewhere?
+      #  If SA processed truncated message from amavis etc, this could also
+      #  be hit legimately..
+      if (!$found_end_boundary) {
+        # TODO: This is duplicate code from few pages down below..
+        while (length ($_) > MAX_BODY_LINE_LENGTH) {
+          push (@{$part_array}, substr($_, 0, MAX_BODY_LINE_LENGTH)."\n");
+          substr($_, 0, MAX_BODY_LINE_LENGTH) = '';
+        }
+        push ( @{$part_array}, $_ );
+      }
       # per rfc 1521, the CRLF before the boundary is part of the boundary:
       # NOTE: The CRLF preceding the encapsulation line is conceptually
       # attached to the boundary so that it is possible to have a part
@@ -810,7 +939,7 @@ sub _parse_multipart {
       # CRLFs preceding the encapsulation line, the first of which is part
       # of the preceding body part, and the second of which is part of the
       # encapsulation boundary.
-      if ($part_array) {
+      elsif ($part_array) {
         chomp( $part_array->[-1] );  # trim the CRLF that's part of the boundary
         splice @{$part_array}, -1 if ( $part_array->[-1] eq '' ); # blank line for the boundary only ...
       }
@@ -835,6 +964,7 @@ sub _parse_multipart {
       # but this causes problems with horizontal lines when the boundary is
       # made up of dashes as well, etc.
       if (defined $boundary) {
+        # no re "strict";  # since perl 5.21.8: Ranges of ASCII printables...
         if ($line =~ /^--\Q${boundary}\E--\s*$/) {
 	  # Make a note that we've seen the end boundary
 	  $self->{mime_boundary_state}->{$boundary}--;
@@ -848,6 +978,13 @@ sub _parse_multipart {
 	}
       }
 
+      # Maximum parts to process
+      if (++$partcnt == 1000) {
+        dbg("message: mimepart limit exceeded, stopping parsing");
+        $self->{'mimepart_limit_exceeded'} = 1;
+        return;
+      }
+
       # make sure we start with a new clean node
       $in_body  = 0;
       $part_msg = Mail::SpamAssassin::Message::Node->new({ normalize=>$self->{normalize} });
@@ -859,6 +996,7 @@ sub _parse_multipart {
 
     if (!$in_body) {
       # s/\s+$//;   # bug 5127: don't clean this up (yet)
+      # no re "strict";  # since perl 5.21.8: Ranges of ASCII printables...
       if (/^[\041-\071\073-\176]+[ \t]*:/) {
         if ($header) {
           my ( $key, $value ) = split ( /:\s*/, $header, 2 );
@@ -940,12 +1078,15 @@ sub _parse_normal {
 
   # attempt to figure out a name for this attachment if there is one ...
   my $disp = $msg->header('content-disposition') || '';
-  if ($disp =~ /name="?([^\";]+)"?/i) {
+  if ($disp =~ /name=\s*"?([^";]+)"?/i) {
     $msg->{'name'} = $1;
   }
   elsif ($ct[3]) {
     $msg->{'name'} = $ct[3];
   }
+  if ($msg->{'name'}) {
+    $msg->{'name'} = Encode::decode("MIME-Header", $msg->{'name'});
+  }
 
   $msg->{'boundary'} = $boundary;
 
@@ -954,109 +1095,103 @@ sub _parse_normal {
   # up RAM with something we're not going to use?
   #
   if ($msg->{'type'} !~ m@^(?:text/(?:plain|html)$|message\b)@) {
-    my $filepath;
-    ($filepath, $msg->{'raw'}) = Mail::SpamAssassin::Util::secure_tmpfile();
-
-    if ($filepath) {
+    my($filepath, $fh);
+    eval {
+      ($filepath, $fh) = Mail::SpamAssassin::Util::secure_tmpfile();  1;
+    } or do {
+      my $eval_stat = $@ ne '' ? $@ : "errno=$!";  chomp $eval_stat;
+      info("message: failed to create a temp file: %s", $eval_stat);
+    };
+    if ($fh) {
       # The temp file was created, add it to the list of pending deletions
       # we cannot just delete immediately in the POSIX idiom, as this is
       # unportable (to win32 at least)
       push @{$self->{tmpfiles}}, $filepath;
-      $msg->{'raw'}->print(@{$body})  or die "error writing to $filepath: $!";
+      dbg("message: storing a message part to file %s", $filepath);
+      $fh->print(@{$body})  or die "error writing to $filepath: $!";
+      $fh->flush  or die "error writing (flush) to $filepath: $!";
+      $msg->{'raw'} = $fh;
     }
   }
 
   # if the part didn't get a temp file, go ahead and store the data in memory
-  if (!exists $msg->{'raw'}) {
+  if (!defined $msg->{'raw'}) {
+    dbg("message: storing a body to memory");
     $msg->{'raw'} = $body;
   }
 }
 
 # ---------------------------------------------------------------------------
 
-sub get_rendered_body_text_array {
+sub get_mimepart_digests {
   my ($self) = @_;
 
-  if (exists $self->{text_rendered}) { return $self->{text_rendered}; }
-
-  $self->{text_rendered} = [];
-
-  # Find all parts which are leaves
-  my @parts = $self->find_parts(qr/./,1);
-  return $self->{text_rendered} unless @parts;
-
-  # the html metadata may have already been set, so let's not bother if it's
-  # already been done.
-  my $html_needs_setting = !exists $self->{metadata}->{html};
-
-  # Go through each part
-  my $text = $self->get_header ('subject') || "\n";
-  for(my $pt = 0 ; $pt <= $#parts ; $pt++ ) {
-    my $p = $parts[$pt];
-
-    # put a blank line between parts ...
-    $text .= "\n";
-
-    my($type, $rnd) = $p->rendered(); # decode this part
-    if ( defined $rnd ) {
-      # Only text/* types are rendered ...
-      $text .= $rnd;
-
-      # TVD - if there are multiple parts, what should we do?
-      # right now, just use the last one.  we may need to give some priority
-      # at some point, ie: use text/html rendered if it exists, or
-      # text/plain rendered as html otherwise.
-      if ($html_needs_setting && $type eq 'text/html') {
-        $self->{metadata}->{html} = $p->{html_results};
-      }
-    }
+  if (!exists $self->{mimepart_digests}) {
+    # traverse all parts which are leaves, recursively
+    $self->{mimepart_digests} =
+      [ map(sha1_hex($_->decode) . ':' . lc($_->{type}||''),
+            $self->find_parts(qr/^/,1,1)) ];
   }
-
-  # whitespace handling (warning: small changes have large effects!)
-  $text =~ s/\n+\s*\n+/\f/gs;		# double newlines => form feed
-  $text =~ tr/ \t\n\r\x0b\xa0/ /s;	# whitespace => space
-  $text =~ tr/\f/\n/;			# form feeds => newline
-  
-  # warn "message: $text";
-
-  my @textary = split_into_array_of_short_lines ($text);
-  $self->{text_rendered} = \@textary;
-
-  return $self->{text_rendered};
+  return $self->{mimepart_digests};
 }
 
 # ---------------------------------------------------------------------------
 
-# TODO: possibly this should just replace get_rendered_body_text_array().
-# (although watch out, this one doesn't copy {html} to metadata)
-sub get_visible_rendered_body_text_array {
-  my ($self) = @_;
+# common code for get_rendered_body_text_array,
+# get_visible_rendered_body_text_array, get_invisible_rendered_body_text_array
+#
+sub get_body_text_array_common {
+  my ($self, $method_name) = @_;
 
-  if (exists $self->{text_visible_rendered}) {
-    return $self->{text_visible_rendered};
-  }
+  my $key = 'text_' . $method_name;
+  if (exists $self->{$key}) { return $self->{$key} }
 
-  $self->{text_visible_rendered} = [];
+  $self->{$key} = [];
+
+  my $scansize = $self->{body_part_scan_size};
 
   # Find all parts which are leaves
   my @parts = $self->find_parts(qr/./,1);
-  return $self->{text_visible_rendered} unless @parts;
+  return $self->{$key} unless @parts;
 
   # the html metadata may have already been set, so let's not bother if it's
   # already been done.
   my $html_needs_setting = !exists $self->{metadata}->{html};
 
+  my $text = $method_name eq 'invisible_rendered' ? ''
+               : ($self->get_header('subject') || "\n");
+
   # Go through each part
-  my $text = $self->get_header ('subject') || "\n";
-  for(my $pt = 0 ; $pt <= $#parts ; $pt++ ) {
+  for (my $pt = 0 ; $pt <= $#parts ; $pt++ ) {
     my $p = $parts[$pt];
 
     # put a blank line between parts ...
-    $text .= "\n";
+    $text .= "\n"  if $text ne '';
 
-    my($type, $rnd) = $p->visible_rendered(); # decode this part
-    if ( defined $rnd ) {
-      # Only text/* types are rendered ...
+    my($type, $rnd) = $p->$method_name();  # decode this part
+    # Only text/* types are rendered ...
+    if (defined $rnd) {
+      # Truncate to body_part_scan_size
+      if ($scansize && length($rnd) > $scansize) {
+        # Try truncating at boundary, nearest 1k block is fine
+        # Newline first
+        my $idx = index($rnd, "\n", $scansize);
+        if ($idx >= 0 && $idx - $scansize <= 1024) {
+          substr($rnd, $idx+1) = ''; # truncate
+        }
+        else {
+          # Space is fine too for rendered
+          $idx = index($rnd, " ", $scansize);
+          if ($idx >= 0 && $idx - $scansize <= 1024) {
+            substr($rnd, $idx+1) = ''; # truncate
+          }
+          else {
+            substr($rnd, $scansize) = ''; # truncate
+          }
+        }
+      }
+
+      # Add to rendered text
       $text .= $rnd;
 
       # TVD - if there are multiple parts, what should we do?
@@ -1071,64 +1206,31 @@ sub get_visible_rendered_body_text_array {
 
   # whitespace handling (warning: small changes have large effects!)
   $text =~ s/\n+\s*\n+/\f/gs;		# double newlines => form feed
-  $text =~ tr/ \t\n\r\x0b\xa0/ /s;	# whitespace => space
+# $text =~ tr/ \t\n\r\x0b\xa0/ /s;	# whitespace (incl. VT, NBSP) => space
+  $text =~ tr/ \t\n\r\x0b/ /s;		# whitespace (incl. VT) => space
   $text =~ tr/\f/\n/;			# form feeds => newline
 
-  my @textary = split_into_array_of_short_lines ($text);
-  $self->{text_visible_rendered} = \@textary;
+  my @textary = split_into_array_of_short_lines($text);
+  $self->{$key} = \@textary;
 
-  return $self->{text_visible_rendered};
+  return $self->{$key};
+}
+
+# ---------------------------------------------------------------------------
+
+sub get_rendered_body_text_array {
+  my ($self) = @_;
+  return $self->get_body_text_array_common('rendered');
+}
+
+sub get_visible_rendered_body_text_array {
+  my ($self) = @_;
+  return $self->get_body_text_array_common('visible_rendered');
 }
 
 sub get_invisible_rendered_body_text_array {
   my ($self) = @_;
-
-  if (exists $self->{text_invisible_rendered}) {
-    return $self->{text_invisible_rendered};
-  }
-
-  $self->{text_invisible_rendered} = [];
-
-  # Find all parts which are leaves
-  my @parts = $self->find_parts(qr/./,1);
-  return $self->{text_invisible_rendered} unless @parts;
-
-  # the html metadata may have already been set, so let's not bother if it's
-  # already been done.
-  my $html_needs_setting = !exists $self->{metadata}->{html};
-
-  # Go through each part
-  my $text = '';
-  for(my $pt = 0 ; $pt <= $#parts ; $pt++ ) {
-    my $p = $parts[$pt];
-
-    # put a blank line between parts ...
-    $text .= "\n" if ( $text );
-
-    my($type, $rnd) = $p->invisible_rendered(); # decode this part
-    if ( defined $rnd ) {
-      # Only text/* types are rendered ...
-      $text .= $rnd;
-
-      # TVD - if there are multiple parts, what should we do?
-      # right now, just use the last one.  we may need to give some priority
-      # at some point, ie: use text/html rendered if it exists, or
-      # text/plain rendered as html otherwise.
-      if ($html_needs_setting && $type eq 'text/html') {
-        $self->{metadata}->{html} = $p->{html_results};
-      }
-    }
-  }
-
-  # whitespace handling (warning: small changes have large effects!)
-  $text =~ s/\n+\s*\n+/\f/gs;		# double newlines => form feed
-  $text =~ tr/ \t\n\r\x0b\xa0/ /s;	# whitespace => space
-  $text =~ tr/\f/\n/;			# form feeds => newline
-
-  my @textary = split_into_array_of_short_lines ($text);
-  $self->{text_invisible_rendered} = \@textary;
-
-  return $self->{text_invisible_rendered};
+  return $self->get_body_text_array_common('invisible_rendered');
 }
 
 # ---------------------------------------------------------------------------
@@ -1139,6 +1241,8 @@ sub get_decoded_body_text_array {
   if (defined $self->{text_decoded}) { return $self->{text_decoded}; }
   $self->{text_decoded} = [ ];
 
+  my $scansize = $self->{rawbody_part_scan_size};
+
   # Find all parts which are leaves
   my @parts = $self->find_parts(qr/^(?:text|message)\b/i,1);
   return $self->{text_decoded} unless @parts;
@@ -1149,9 +1253,38 @@ sub get_decoded_body_text_array {
     # and not displayed
     next if ($parts[$pt]->{'type'} eq 'text/calendar');
 
-    push(@{$self->{text_decoded}}, "\n") if ( @{$self->{text_decoded}} );
+    my $rawbody = $parts[$pt]->decode();
+    my $rawlen = length($rawbody);
+
+    # Truncate to rawbody_part_scan_size
+    if ($scansize && $rawlen > $scansize) {
+      # Try truncating at boundary, nearest 1k block is fine
+      # Newline first
+      my $idx = index($rawbody, "\n", $scansize);
+      if ($idx >= 0 && $idx - $scansize <= 1024) {
+        substr($rawbody, $idx+1) = ''; # truncate
+      }
+      else {
+        # Try > next, might be html
+        $idx = index($rawbody, '>', $scansize);
+        if ($idx >= 0 && $idx - $scansize <= 1024) {
+          substr($rawbody, $idx+1) = ''; # truncate
+        }
+        else {
+          # Space as last effort
+          $idx = index($rawbody, ' ', $scansize);
+          if ($idx >= 0 && $idx - $scansize <= 1024) {
+            substr($rawbody, $idx+1) = ''; # truncate
+          }
+          else {
+            substr($rawbody, $scansize) = ''; # truncate
+          }
+        }
+      }
+    }
+
     push(@{$self->{text_decoded}},
-         split_into_array_of_short_paragraphs($parts[$pt]->decode()));
+         split_into_array_of_short_paragraphs($rawbody));
   }
 
   return $self->{text_decoded};
@@ -1163,7 +1296,7 @@ sub split_into_array_of_short_lines {
   my @result;
   foreach my $line (split (/^/m, $_[0])) {
     while (length ($line) > MAX_BODY_LINE_LENGTH) {
-      # try splitting "nicely" so that we don't chop an url in half or
+      # try splitting "nicely" so that we don't chop a url in half or
       # something.  if there's no space, then just split at max length.
       my $length = rindex($line, ' ', MAX_BODY_LINE_LENGTH) + 1;
       $length ||= MAX_BODY_LINE_LENGTH;
@@ -1176,19 +1309,25 @@ sub split_into_array_of_short_lines {
 
 # ---------------------------------------------------------------------------
 
-# split a text into array of paragraphs of sizes between
-# $chunk_size and 2 * $chunk_size, returning the resulting array
+# Split a text into array of paragraphs of sizes between
+# 1-2 * MAX_BODY_LINE_LENGTH, returning the resulting array.
+# Mainly used for rawbody splitting purposes
 
 sub split_into_array_of_short_paragraphs {
-  my @result;
-  my $chunk_size = 1024;
   my $text_l = length($_[0]);
-  my($j,$ofs);
-  for ($ofs = 0;  $text_l - $ofs > 2 * $chunk_size;  $ofs = $j+1) {
-    $j = index($_[0], "\n", $ofs+$chunk_size);
-    if ($j < 0) {
-      $j = index($_[0], " ", $ofs+$chunk_size);
-      if ($j < 0) { $j = $ofs+$chunk_size }
+  return ($_[0]) if $text_l <= MAX_BODY_LINE_LENGTH;
+  my(@result,$j,$ofs);
+  # Try to split in order \n, > (html?), space
+  for ($ofs = 0; $text_l - $ofs > 2 * MAX_BODY_LINE_LENGTH; $ofs = $j+1) {
+    $j = index($_[0], "\n", $ofs + MAX_BODY_LINE_LENGTH);
+    if ($j < 0 || $j-$ofs+1 > 2 * MAX_BODY_LINE_LENGTH) {
+      $j = index($_[0], ">", $ofs + MAX_BODY_LINE_LENGTH);
+      if ($j < 0 || $j-$ofs+1 > 2 * MAX_BODY_LINE_LENGTH) {
+        $j = index($_[0], " ", $ofs + MAX_BODY_LINE_LENGTH);
+        if ($j < 0 || $j-$ofs+1 > 2 * MAX_BODY_LINE_LENGTH) {
+          $j = $ofs + MAX_BODY_LINE_LENGTH;
+        }
+      }
     }
     push(@result, substr($_[0], $ofs, $j-$ofs+1));
   }
