@@ -60,15 +60,14 @@ package Mail::SpamAssassin::Plugin::AWL;
 
 use strict;
 use warnings;
-use bytes;
+# use bytes;
 use re 'taint';
 use Mail::SpamAssassin::Plugin;
 use Mail::SpamAssassin::AutoWhitelist;
 use Mail::SpamAssassin::Util qw(untaint_var);
 use Mail::SpamAssassin::Logger;
 
-use vars qw(@ISA);
-@ISA = qw(Mail::SpamAssassin::Plugin);
+our @ISA = qw(Mail::SpamAssassin::Plugin);
 
 # constructor: register the eval rule
 sub new {
@@ -326,7 +325,7 @@ The file mode bits used for the automatic-whitelist directory or file.
 
 Make sure you specify this using the 'x' mode bits set, as it may also be used
 to create directories.  However, if a file is created, the resulting file will
-not have any execute bits set (the umask is set to 111).
+not have any execute bits set (the umask is set to 0111).
 
 =cut
 
@@ -337,7 +336,7 @@ not have any execute bits set (the umask is set to 111).
 		type => $Mail::SpamAssassin::Conf::CONF_TYPE_NUMERIC,
 		code => sub {
 		  my ($self, $key, $value, $line) = @_;
-		  if ($value !~ /^0?\d{3}$/) {
+		  if ($value !~ /^0?[0-7]{3}$/) {
                     return $Mail::SpamAssassin::Conf::INVALID_VALUE;
                   }
 		  $self->{auto_whitelist_file_mode} = untaint_var($value);
@@ -437,6 +436,7 @@ sub check_from_in_auto_whitelist {
       # or if the test is marked as "noautolearn"
       next if !$scores->{$test};
       next if exists $tflags->{$test} && $tflags->{$test} =~ /\bnoautolearn\b/;
+      return 0 if exists $tflags->{$test} && $tflags->{$test} =~ /\bnoawl\b/;
       $points += $scores->{$test};
     }
 
