@@ -45,7 +45,7 @@ package Mail::SpamAssassin::Plugin::Bayes;
 
 use strict;
 use warnings;
-use bytes;
+# use bytes;
 use re 'taint';
 
 BEGIN {
@@ -65,20 +65,13 @@ use Mail::SpamAssassin::Bayes::CombineChi;
 
 our @ISA = qw(Mail::SpamAssassin::Plugin);
 
-use vars qw{
-  $IGNORED_HDRS
-  $MARK_PRESENCE_ONLY_HDRS
-  %HEADER_NAME_COMPRESSION
-  $OPPORTUNISTIC_LOCK_VALID
-};
-
 # Which headers should we scan for tokens?  Don't use all of them, as it's easy
 # to pick up spurious clues from some.  What we now do is use all of them
 # *less* these well-known headers; that way we can pick up spammers' tracking
 # headers (which are obviously not well-known in advance!).
 
 # Received is handled specially
-$IGNORED_HDRS = qr{(?: (?:X-)?Sender    # misc noise
+our $IGNORED_HDRS = qr{(?: (?:X-)?Sender    # misc noise
   |Delivered-To |Delivery-Date
   |(?:X-)?Envelope-To
   |X-MIME-Auto[Cc]onverted |X-Converted-To-Plain-Text
@@ -101,6 +94,7 @@ $IGNORED_HDRS = qr{(?: (?:X-)?Sender    # misc noise
   |X-eGroups-(?:Return|From)
   |X-MDMailing-List
   |X-XEmacs-List
+  |X-Sympa-To
 
   # gatewayed through mailing list (thanks to Allen Smith)
   |(?:X-)?Resent-(?:From|To|Date)
@@ -145,18 +139,21 @@ $IGNORED_HDRS = qr{(?: (?:X-)?Sender    # misc noise
 
 # Note only the presence of these headers, in order to reduce the
 # hapaxen they generate.
-$MARK_PRESENCE_ONLY_HDRS = qr{(?: X-Face
+our $MARK_PRESENCE_ONLY_HDRS = qr{(?: X-Face
   |X-(?:Gnu-?PG|PGP|GPG)(?:-Key)?-Fingerprint
   |D(?:KIM|omainKey)-Signature
 )}ix;
 
-# tweaks tested as of Nov 18 2002 by jm: see SpamAssassin-devel list archives
+# tweaks tested as of Nov 18 2002 by jm posted to -devel at
+# http://sourceforge.net/p/spamassassin/mailman/message/12977556/
 # for results.  The winners are now the default settings.
 use constant IGNORE_TITLE_CASE => 1;
-use constant TOKENIZE_LONG_8BIT_SEQS_AS_TUPLES => 1;
+use constant TOKENIZE_LONG_8BIT_SEQS_AS_TUPLES => 0;
+use constant TOKENIZE_LONG_8BIT_SEQS_AS_UTF8_CHARS => 1;
 use constant TOKENIZE_LONG_TOKENS_AS_SKIPS => 1;
 
-# tweaks of May 12 2003, see SpamAssassin-devel archives again.
+# tweaks by jm on May 12 2003, see -devel email at
+# http://sourceforge.net/p/spamassassin/mailman/message/14844556/
 use constant PRE_CHEW_ADDR_HEADERS => 1;
 use constant CHEW_BODY_URIS => 1;
 use constant CHEW_BODY_MAILADDRS => 1;
@@ -182,7 +179,7 @@ use constant ADD_INVIZ_TOKENS_NO_PREFIX => 0;
 # be 2-way (ie. LHSes that map to the same RHS are not a problem), but mixing
 # tokens from multiple different headers may impact accuracy, so might as well
 # avoid this if possible. These are the top ones from my corpus, BTW (jm).
-%HEADER_NAME_COMPRESSION = (
+our %HEADER_NAME_COMPRESSION = (
   'Message-Id'		=> '*m',
   'Message-ID'		=> '*M',
   'Received'		=> '*r',
@@ -197,13 +194,15 @@ use constant ADD_INVIZ_TOKENS_NO_PREFIX => 0;
   'X-Authentication-Warning' => '*a',
   'Organization'	=> '*o',
   'Organisation'        => '*o',
-  'Content-Type'	=> '*c',
+  'Content-Type'	=> '*ct',
+  'Content-Disposition'	=> '*cd',
+  'Content-Transfer-Encoding' => '*ce',
   'x-spam-relays-trusted' => '*RT',
   'x-spam-relays-untrusted' => '*RU',
 );
 
 # How many seconds should the opportunistic_expire lock be valid?
-$OPPORTUNISTIC_LOCK_VALID = 300;
+our $OPPORTUNISTIC_LOCK_VALID = 300;
 
 # Should we use the Robinson f(w) equation from
 # http://radio.weblogs.com/0101454/stories/2002/09/16/spamDetection.html ?
@@ -249,6 +248,8 @@ sub finish {
   %{$self} = ();
 }
 
+###########################################################################
+
 # Plugin hook.
 # Return this implementation object, for callers that need to know
 # it.  TODO: callers shouldn't *need* to know it! 
@@ -258,11 +259,36 @@ sub learner_get_implementation { return shift; }
 
 ###########################################################################
 
+# Plugin hook.
+# Called in the parent process shortly before forking off child processes.
+sub prefork_init {
+  my ($self) = @_;
+
+  if ($self->{store} && $self->{store}->UNIVERSAL::can('prefork_init')) {
+    $self->{store}->prefork_init;
+  }
+}
+
+###########################################################################
+
+# Plugin hook.
+# Called in a child process shortly after being spawned.
+sub spamd_child_init {
+  my ($self) = @_;
+
+  if ($self->{store} && $self->{store}->UNIVERSAL::can('spamd_child_init')) {
+    $self->{store}->spamd_child_init;
+  }
+}
+
+###########################################################################
+
+# Plugin hook.
 sub check_bayes {
   my ($self, $pms, $fulltext, $min, $max) = @_;
 
-  return 0 if (!$pms->{conf}->{use_learner});
-  return 0 if (!$pms->{conf}->{use_bayes} || !$pms->{conf}->{use_bayes_rules});
+  return 0 if (!$self->{conf}->{use_learner});
+  return 0 if (!$self->{conf}->{use_bayes} || !$self->{conf}->{use_bayes_rules});
 
   if (!exists ($pms->{bayes_score})) {
     my $timer = $self->{main}->time_method("check_bayes");
@@ -273,7 +299,7 @@ sub check_bayes {
       ($min == 0 || $pms->{bayes_score} > $min) &&
       ($max eq "undef" || $pms->{bayes_score} <= $max))
   {
-      if ($pms->{conf}->{detailed_bayes_score}) {
+      if ($self->{conf}->{detailed_bayes_score}) {
         $pms->test_log(sprintf ("score: %3.4f, hits: %s",
                                  $pms->{bayes_score},
                                  $pms->{bayes_hits}));
@@ -294,7 +320,7 @@ sub learner_close {
   my ($self, $params) = @_;
   my $quiet = $params->{quiet};
 
-  # do a sanity check here.  Wierd things happen if we remain tied
+  # do a sanity check here.  Weird things happen if we remain tied
   # after compiling; for example, spamd will never see that the
   # number of messages has reached the bayes-scanning threshold.
   if ($self->{store}->db_readable()) {
@@ -349,6 +375,7 @@ sub learn_message {
 
   eval {
     local $SIG{'__DIE__'};	# do not run user die() traps in here
+    my $timer = $self->{main}->time_method("b_learn");
 
     my $ok;
     if ($self->{main}->{learn_to_journal}) {
@@ -386,25 +413,25 @@ sub _learn_trapped {
     @msgid = $self->get_msgid($msg);
   }
 
-  foreach $msgid ( @msgid ) {
-    my $seen = $self->{store}->seen_get ($msgid);
+  foreach my $msgid_t ( @msgid ) {
+    my $seen = $self->{store}->seen_get ($msgid_t);
 
     if (defined ($seen)) {
       if (($seen eq 's' && $isspam) || ($seen eq 'h' && !$isspam)) {
-        dbg("bayes: $msgid already learnt correctly, not learning twice");
+        dbg("bayes: $msgid_t already learnt correctly, not learning twice");
         return 0;
       } elsif ($seen !~ /^[hs]$/) {
-        warn("bayes: db_seen corrupt: value='$seen' for $msgid, ignored");
+        warn("bayes: db_seen corrupt: value='$seen' for $msgid_t, ignored");
       } else {
         # bug 3704: If the message was already learned, don't try learning it again.
         # this prevents, for instance, manually learning as spam, then autolearning
         # as ham, or visa versa.
         if ($self->{main}->{learn_no_relearn}) {
-	  dbg("bayes: $msgid already learnt as opposite, not re-learning");
+	  dbg("bayes: $msgid_t already learnt as opposite, not re-learning");
 	  return 0;
 	}
 
-        dbg("bayes: $msgid already learnt as opposite, forgetting first");
+        dbg("bayes: $msgid_t already learnt as opposite, forgetting first");
 
         # kluge so that forget() won't untie the db on us ...
         my $orig = $self->{main}->{learn_caller_will_untie};
@@ -431,12 +458,6 @@ sub _learn_trapped {
   # Now that we're sure we haven't seen this message before ...
   $msgid = $msgid[0];
 
-  if ($isspam) {
-    $self->{store}->nspam_nham_change (1, 0);
-  } else {
-    $self->{store}->nspam_nham_change (0, 1);
-  }
-
   my $msgatime = $msg->receive_date();
 
   # If the message atime comes back as being more than 1 day in the
@@ -447,10 +468,14 @@ sub _learn_trapped {
 
   my $tokens = $self->tokenize($msg, $msgdata);
 
-  if ($isspam) {
-    $self->{store}->multi_tok_count_change(1, 0, $tokens, $msgatime);
-  } else {
-    $self->{store}->multi_tok_count_change(0, 1, $tokens, $msgatime);
+  { my $timer = $self->{main}->time_method('b_count_change');
+    if ($isspam) {
+      $self->{store}->nspam_nham_change(1, 0);
+      $self->{store}->multi_tok_count_change(1, 0, $tokens, $msgatime);
+    } else {
+      $self->{store}->nspam_nham_change(0, 1);
+      $self->{store}->multi_tok_count_change(0, 1, $tokens, $msgatime);
+    }
   }
 
   $self->{store}->seen_put ($msgid, ($isspam ? 's' : 'h'));
@@ -484,6 +509,7 @@ sub forget_message {
   # synchronously
   eval {
     local $SIG{'__DIE__'};	# do not run user die() traps in here
+    my $timer = $self->{main}->time_method("b_learn");
 
     my $ok;
     if ($self->{main}->{learn_to_journal}) {
@@ -666,19 +692,27 @@ sub scan {
 
   dbg("bayes: corpus size: nspam = $ns, nham = $nn");
 
-  my $msgdata = $self->_get_msgdata_from_permsgstatus ($permsgstatus);
+  my $msgtokens;
+  { my $timer = $self->{main}->time_method('b_tokenize');
+    my $msgdata = $self->_get_msgdata_from_permsgstatus ($permsgstatus);
+    $msgtokens = $self->tokenize($msg, $msgdata);
+  }
 
-  my $msgtokens = $self->tokenize($msg, $msgdata);
+  my $tokensdata;
+  { my $timer = $self->{main}->time_method('b_tok_get_all');
+    $tokensdata = $self->{store}->tok_get_all(keys %{$msgtokens});
+  }
 
-  my $tokensdata = $self->{store}->tok_get_all(keys %{$msgtokens});
+  my $timer_compute_prob = $self->{main}->time_method('b_comp_prob');
+
+  my $probabilities_ref =
+    $self->_compute_prob_for_all_tokens($tokensdata, $ns, $nn);
 
   my %pw;
-
   foreach my $tokendata (@{$tokensdata}) {
-    my ($token, $tok_spam, $tok_ham, $atime) = @{$tokendata};
-    my $prob = $self->_compute_prob_for_token($token, $ns, $nn, $tok_spam, $tok_ham);
+    my $prob = shift(@$probabilities_ref);
     next unless defined $prob;
-
+    my ($token, $tok_spam, $tok_ham, $atime) = @{$tokendata};
     $pw{$token} = {
       prob => $prob,
       spam_count => $tok_spam,
@@ -687,15 +721,17 @@ sub scan {
     };
   }
 
+  my @pw_keys = keys %pw;
+
   # If none of the tokens were found in the DB, we're going to skip
   # this message...
-  if (!keys %pw) {
+  if (!@pw_keys) {
     dbg("bayes: cannot use bayes on this message; none of the tokens were found in the database");
     goto skip;
   }
 
   my $tcount_total = keys %{$msgtokens};
-  my $tcount_learned = keys %pw;
+  my $tcount_learned = scalar @pw_keys;
 
   # Figure out the message receive time (used as atime below)
   # If the message atime comes back as being in the future, something's
@@ -705,50 +741,47 @@ sub scan {
   my $now = time;
   $msgatime = $now if ( $msgatime > $now );
 
-  # now take the $count most significant tokens and calculate probs using
-  # Robinson's formula.
-  my $count = N_SIGNIFICANT_TOKENS;
-  my @sorted;
-
   my @touch_tokens;
   my $tinfo_spammy = $permsgstatus->{bayes_token_info_spammy} = [];
   my $tinfo_hammy = $permsgstatus->{bayes_token_info_hammy} = [];
 
-  my %tok_strength = map { $_ => (abs($pw{$_}->{prob} - 0.5)) } keys %pw;
+  my %tok_strength = map( ($_, abs($pw{$_}->{prob} - 0.5)), @pw_keys);
   my $log_each_token = (would_log('dbg', 'bayes') > 1);
 
-  foreach my $tok (sort {
-              $tok_strength{$b} <=> $tok_strength{$a}
-            } keys %pw)
-  {
-    if ($count-- < 0) { last; }
-    next if ($tok_strength{$tok} <
-                $Mail::SpamAssassin::Bayes::Combine::MIN_PROB_STRENGTH);
+  # now take the most significant tokens and calculate probs using
+  # Robinson's formula.
 
-    my $pw = $pw{$tok}->{prob};
+  @pw_keys = sort { $tok_strength{$b} <=> $tok_strength{$a} } @pw_keys;
+
+  if (@pw_keys > N_SIGNIFICANT_TOKENS) { $#pw_keys = N_SIGNIFICANT_TOKENS - 1 }
+
+  my @sorted;
+  foreach my $tok (@pw_keys) {
+    next if $tok_strength{$tok} <
+                $Mail::SpamAssassin::Bayes::Combine::MIN_PROB_STRENGTH;
+
+    my $pw_tok = $pw{$tok};
+    my $pw_prob = $pw_tok->{prob};
 
     # What's more expensive, scanning headers for HAMMYTOKENS and
     # SPAMMYTOKENS tags that aren't there or collecting data that
     # won't be used?  Just collecting the data is certainly simpler.
     #
     my $raw_token = $msgtokens->{$tok} || "(unknown)";
-    my $s = $pw{$tok}->{spam_count};
-    my $n = $pw{$tok}->{ham_count};
-    my $a = $pw{$tok}->{atime};
+    my $s = $pw_tok->{spam_count};
+    my $n = $pw_tok->{ham_count};
+    my $a = $pw_tok->{atime};
 
-    if ($pw < 0.5) {
-      push @$tinfo_hammy,  [$raw_token,$pw,$s,$n,$a];
-    } else {
-      push @$tinfo_spammy, [$raw_token,$pw,$s,$n,$a];
-    }
+    push( @{ $pw_prob < 0.5 ? $tinfo_hammy : $tinfo_spammy },
+          [$raw_token, $pw_prob, $s, $n, $a] );
 
-    push (@sorted, $pw);
+    push(@sorted, $pw_prob);
 
     # update the atime on this token, it proved useful
     push(@touch_tokens, $tok);
 
     if ($log_each_token) {
-      dbg("bayes: token '$raw_token' => $pw");
+      dbg("bayes: token '$raw_token' => $pw_prob");
     }
   }
 
@@ -760,6 +793,7 @@ sub scan {
   }
 
   $score = Mail::SpamAssassin::Bayes::Combine::combine($ns, $nn, \@sorted);
+  undef $timer_compute_prob;  # end a timing section
 
   # Couldn't come up with a probability?
   goto skip unless defined $score;
@@ -769,7 +803,12 @@ sub scan {
   # no need to call tok_touch_all unless there were significant
   # tokens and a score was returned
   # we don't really care about the return value here
-  $self->{store}->tok_touch_all(\@touch_tokens, $msgatime);
+
+  { my $timer = $self->{main}->time_method('b_tok_touch_all');
+    $self->{store}->tok_touch_all(\@touch_tokens, $msgatime);
+  }
+
+  my $timer_finish = $self->{main}->time_method('b_finish');
 
   $permsgstatus->{bayes_nspam} = $ns;
   $permsgstatus->{bayes_nham} = $nn;
@@ -790,6 +829,11 @@ skip:
     dbg("bayes: not scoring message, returning undef");
   }
 
+  undef $timer_compute_prob;  # end a timing section if still running
+  if (!defined $timer_finish) {
+    $timer_finish = $self->{main}->time_method('b_finish');
+  }
+
   # Take any opportunistic actions we can take
   if ($self->{main}->{opportunistic_expire_check_only}) {
     # we're supposed to report on expiry only -- so do the
@@ -820,27 +864,30 @@ skip:
   $permsgstatus->set_tag ('BAYESTC', $tcount_total);
 
   $permsgstatus->set_tag ('HAMMYTOKENS', sub {
+              my $pms = shift;
               $self->bayes_report_make_list
-                ($permsgstatus, $permsgstatus->{bayes_token_info_hammy}, shift);
+                ($pms, $pms->{bayes_token_info_hammy}, shift);
             });
 
   $permsgstatus->set_tag ('SPAMMYTOKENS', sub {
+              my $pms = shift;
               $self->bayes_report_make_list
-                ($permsgstatus, $permsgstatus->{bayes_token_info_spammy}, shift);
+                ($pms, $pms->{bayes_token_info_spammy}, shift);
             });
 
   $permsgstatus->set_tag ('TOKENSUMMARY', sub {
-              if ( defined $permsgstatus->{tag_data}{BAYESTC} )
+              my $pms = shift;
+              if ( defined $pms->{tag_data}{BAYESTC} )
                 {
-                  my $tcount_neutral = $permsgstatus->{tag_data}{BAYESTCLEARNED}
-                                    - $permsgstatus->{tag_data}{BAYESTCSPAMMY}
-                                    - $permsgstatus->{tag_data}{BAYESTCHAMMY};
-                  my $tcount_new = $permsgstatus->{tag_data}{BAYESTC}
-                                    - $permsgstatus->{tag_data}{BAYESTCLEARNED};
+                  my $tcount_neutral = $pms->{tag_data}{BAYESTCLEARNED}
+                                     - $pms->{tag_data}{BAYESTCSPAMMY}
+                                     - $pms->{tag_data}{BAYESTCHAMMY};
+                  my $tcount_new = $pms->{tag_data}{BAYESTC}
+                                 - $pms->{tag_data}{BAYESTCLEARNED};
                   "Tokens: new, $tcount_new; "
-                    ."hammy, $permsgstatus->{tag_data}{BAYESTCHAMMY}; "
+                    ."hammy, $pms->{tag_data}{BAYESTCHAMMY}; "
                     ."neutral, $tcount_neutral; "
-                    ."spammy, $permsgstatus->{tag_data}{BAYESTCSPAMMY}."
+                    ."spammy, $pms->{tag_data}{BAYESTCSPAMMY}."
                 } else {
                   "Bayes not run.";
                 }
@@ -930,24 +977,32 @@ sub get_msgid {
     push(@msgid, $msgid);
   }
 
-  # Use sha1_hex(Date:, last received: and top N bytes of body)
+  # Modified 2012-01-17  per bug 5185 to remove last received from msg_id calculation
+
+  # Use sha1_hex(Date: and top N bytes of body)
   # where N is MIN(1024 bytes, 1/2 of body length)
   #
   my $date = $msg->get_header("Date");
   $date = "None" if (!defined $date || $date eq ''); # No Date?
 
-  my @rcvd = $msg->get_header("Received");
-  my $rcvd = $rcvd[$#rcvd];
-  $rcvd = "None" if (!defined $rcvd || $rcvd eq ''); # No Received?
+  #Removed per bug 5185
+  #my @rcvd = $msg->get_header("Received");
+  #my $rcvd = $rcvd[$#rcvd];
+  #$rcvd = "None" if (!defined $rcvd || $rcvd eq ''); # No Received?
 
   # Make a copy since pristine_body is a reference ...
   my $body = join('', $msg->get_pristine_body());
+
   if (length($body) > 64) { # Small Body?
     my $keep = ( length $body > 2048 ? 1024 : int(length($body) / 2) );
     substr($body, $keep) = '';
   }
 
-  unshift(@msgid, sha1_hex($date."\000".$rcvd."\000".$body).'@sa_generated');
+  #Stripping all CR and LF so that testing midstream from MTA and post delivery don't 
+  #generate different id's simply because of LF<->CR<->CRLF changes.
+  $body =~ s/[\r\n]//g;
+
+  unshift(@msgid, sha1_hex($date."\000".$body).'@sa_generated');
 
   return wantarray ? @msgid : $msgid[0];
 }
@@ -977,12 +1032,18 @@ sub get_body_from_msg {
 }
 
 sub _get_msgdata_from_permsgstatus {
-  my ($self, $msg) = @_;
+  my ($self, $pms) = @_;
 
+  my $t_src = $self->{conf}->{bayes_token_sources};
   my $msgdata = { };
-  $msgdata->{bayes_token_body} = $msg->{msg}->get_visible_rendered_body_text_array();
-  $msgdata->{bayes_token_inviz} = $msg->{msg}->get_invisible_rendered_body_text_array();
-  @{$msgdata->{bayes_token_uris}} = $msg->get_uri_list();
+  $msgdata->{bayes_token_body} =
+    $pms->{msg}->get_visible_rendered_body_text_array() if $t_src->{visible};
+  $msgdata->{bayes_token_inviz} =
+    $pms->{msg}->get_invisible_rendered_body_text_array() if $t_src->{invisible};
+  $msgdata->{bayes_mimepart_digests} =
+    $pms->{msg}->get_mimepart_digests() if $t_src->{mimepart};
+  @{$msgdata->{bayes_token_uris}} =
+    $pms->get_uri_list() if $t_src->{uri};
   return $msgdata;
 }
 
@@ -992,36 +1053,71 @@ sub _get_msgdata_from_permsgstatus {
 sub tokenize {
   my ($self, $msg, $msgdata) = @_;
 
-  # the body
-  my @tokens = map { $self->_tokenize_line ($_, '', 1) }
-                                    @{$msgdata->{bayes_token_body}};
+  my $t_src = $self->{conf}->{bayes_token_sources};
+  my @tokens;
 
-  # the URI list
-  push (@tokens, map { $self->_tokenize_line ($_, '', 2) }
-                                    @{$msgdata->{bayes_token_uris}});
-
-  # add invisible tokens
-  if (ADD_INVIZ_TOKENS_I_PREFIX) {
-    push (@tokens, map { $self->_tokenize_line ($_, "I*:", 1) }
-                                    @{$msgdata->{bayes_token_inviz}});
+  # visible tokens from the body
+  if ($msgdata->{bayes_token_body}) {
+    my(@t) = map($self->_tokenize_line ($_, '', 1),
+                 @{$msgdata->{bayes_token_body}} );
+    dbg("bayes: tokenized body: %d tokens", scalar @t);
+    push(@tokens, @t);
   }
-  if (ADD_INVIZ_TOKENS_NO_PREFIX) {
-    push (@tokens, map { $self->_tokenize_line ($_, "", 1) }
-                                    @{$msgdata->{bayes_token_inviz}});
+  # the URI list
+  if ($msgdata->{bayes_token_uris}) {
+    my(@t) = map($self->_tokenize_line ($_, '', 2),
+                 @{$msgdata->{bayes_token_uris}} );
+    dbg("bayes: tokenized uri: %d tokens", scalar @t);
+    push(@tokens, @t);
+  }
+  # add invisible tokens
+  if ($msgdata->{bayes_token_inviz}) {
+    my $tokprefix;
+    if (ADD_INVIZ_TOKENS_I_PREFIX)  { $tokprefix = 'I*:' }
+    if (ADD_INVIZ_TOKENS_NO_PREFIX) { $tokprefix = '' }
+    if (defined $tokprefix) {
+      my(@t) = map($self->_tokenize_line ($_, $tokprefix, 1),
+                   @{$msgdata->{bayes_token_inviz}} );
+      dbg("bayes: tokenized invisible: %d tokens", scalar @t);
+      push(@tokens, @t);
+    }
+  }
+
+  # add digests and Content-Type of all MIME parts
+  if ($msgdata->{bayes_mimepart_digests}) {
+    my %shorthand = (  # some frequent MIME part contents for human readability
+     'da39a3ee5e6b4b0d3255bfef95601890afd80709:text/plain'=> 'Empty-Plaintext',
+     'da39a3ee5e6b4b0d3255bfef95601890afd80709:text/html' => 'Empty-HTML',
+     'da39a3ee5e6b4b0d3255bfef95601890afd80709:text/xml'  => 'Empty-XML',
+     'adc83b19e793491b1c6ea0fd8b46cd9f32e592fc:text/plain'=> 'OneNL-Plaintext',
+     'adc83b19e793491b1c6ea0fd8b46cd9f32e592fc:text/html' => 'OneNL-HTML',
+     '71853c6197a6a7f222db0f1978c7cb232b87c5ee:text/plain'=> 'TwoNL-Plaintext',
+     '71853c6197a6a7f222db0f1978c7cb232b87c5ee:text/html' => 'TwoNL-HTML',
+    );
+    my(@t) = map('MIME:' . ($shorthand{$_} || $_),
+                 @{ $msgdata->{bayes_mimepart_digests} });
+    dbg("bayes: tokenized mime parts: %d tokens", scalar @t);
+    dbg("bayes: mime-part token %s", $_) for @t;
+    push(@tokens, @t);
   }
 
   # Tokenize the headers
-  my %hdrs = $self->_tokenize_headers ($msg);
-  while( my($prefix, $value) = each %hdrs ) {
-    push(@tokens, $self->_tokenize_line ($value, "H$prefix:", 0));
+  if ($t_src->{header}) {
+    my(@t);
+    my %hdrs = $self->_tokenize_headers ($msg);
+    while( my($prefix, $value) = each %hdrs ) {
+      push(@t, $self->_tokenize_line ($value, "H$prefix:", 0));
+    }
+    dbg("bayes: tokenized header: %d tokens", scalar @t);
+    push(@tokens, @t);
   }
 
   # Go ahead and uniq the array, skip null tokens (can happen sometimes)
   # generate an SHA1 hash and take the lower 40 bits as our token
   my %tokens;
   foreach my $token (@tokens) {
-    next unless length($token); # skip 0 length tokens
-    $tokens{substr(sha1($token), -5)} = $token;
+  # dbg("bayes: token: %s", $token);
+    $tokens{substr(sha1($token), -5)} = $token  if $token ne '';
   }
 
   # return the keys == tokens ...
@@ -1039,7 +1135,17 @@ sub _tokenize_line {
   # include quotes, .'s and -'s for URIs, and [$,]'s for Nigerian-scam strings,
   # and ISO-8859-15 alphas.  Do not split on @'s; better results keeping it.
   # Some useful tokens: "$31,000,000" "www.clock-speed.net" "f*ck" "Hits!"
-  tr/-A-Za-z0-9,\@\*\!_'"\$.\241-\377 / /cs;
+
+  ### (previous:)  tr/-A-Za-z0-9,\@\*\!_'"\$.\241-\377 / /cs;
+
+  ### (now): see Bug 7130 for rationale (slower, but makes UTF-8 chars atomic)
+  s{ ( [A-Za-z0-9,@*!_'"\$. -]+  |
+       [\xC0-\xDF][\x80-\xBF]    |
+       [\xE0-\xEF][\x80-\xBF]{2} |
+       [\xF0-\xF4][\x80-\xBF]{3} |
+       [\xA1-\xFF] ) | . }
+   { defined $1 ? $1 : ' ' }xsge;
+  # should we also turn NBSP ( \xC2\xA0 ) into space?
 
   # DO split on "..." or "--" or "---"; common formatting error resulting in
   # hapaxes.  Keep the separator itself as a token, though, as long ones can
@@ -1057,6 +1163,10 @@ sub _tokenize_line {
 
   my $magic_re = $self->{store}->get_magic_re();
 
+  # Note that split() in scope of 'use bytes' results in words with utf8 flag
+  # cleared, even if the source string has perl characters semantics !!!
+  # Is this really still desirable?
+
   foreach my $token (split) {
     $token =~ s/^[-'"\.,]+//;        # trim non-alphanum chars at start or end
     $token =~ s/[-'"\.,]+$//;        # so we don't get loads of '"foo' tokens
@@ -1096,12 +1206,24 @@ sub _tokenize_line {
     # the domain ".net" appeared in the To header.
     #
     if ($len > MAX_TOKEN_LENGTH && $token !~ /\*/) {
+
+      if (TOKENIZE_LONG_8BIT_SEQS_AS_UTF8_CHARS && $token =~ /[\x80-\xBF]{2}/) {
+	# Bug 7135
+	# collect 3- and 4-byte UTF-8 sequences, ignore 2-byte sequences
+	my(@t) = $token =~ /( (?: [\xE0-\xEF] | [\xF0-\xF4][\x80-\xBF] )
+                              [\x80-\xBF]{2} )/xsg;
+	if (@t) {
+          push (@rettokens, map($tokprefix.'u8:'.$_, @t));
+	  next;
+	}
+      }
+
       if (TOKENIZE_LONG_8BIT_SEQS_AS_TUPLES && $token =~ /[\xa0-\xff]{2}/) {
 	# Matt sez: "Could be asian? Autrijus suggested doing character ngrams,
 	# but I'm doing tuples to keep the dbs small(er)."  Sounds like a plan
 	# to me! (jm)
 	while ($token =~ s/^(..?)//) {
-	  push (@rettokens, "8:$1");
+	  push (@rettokens, $tokprefix.'8:'.$1);
 	}
 	next;
       }
@@ -1111,10 +1233,18 @@ sub _tokenize_line {
             || ($region == 2 && URIS_TOKENIZE_LONG_TOKENS_AS_SKIPS))
       {
 	# if (TOKENIZE_LONG_TOKENS_AS_SKIPS)
-	# Spambayes trick via Matt: Just retain 7 chars.  Do not retain
-	# the length, it does not help; see my mail to -devel of Nov 20 2002.
+	# Spambayes trick via Matt: Just retain 7 chars.  Do not retain the
+	# length, it does not help; see jm's mail to -devel on Nov 20 2002 at
+	# http://sourceforge.net/p/spamassassin/mailman/message/12977605/
 	# "sk:" stands for "skip".
-	$token = "sk:".substr($token, 0, 7);
+	# Bug 7141: retain seven UTF-8 chars (or other bytes),
+	# if followed by at least two bytes
+	$token =~ s{ ^ ( (?> (?: [\x00-\x7F\xF5-\xFF]      |
+	                         [\xC0-\xDF][\x80-\xBF]    |
+	                         [\xE0-\xEF][\x80-\xBF]{2} |
+	                         [\xF0-\xF4][\x80-\xBF]{3} | . ){7} ))
+	             .{2,} \z }{sk:$1}xs;
+	## (was:)  $token = "sk:".substr($token, 0, 7);  # seven bytes
       }
     }
 
@@ -1352,6 +1482,63 @@ sub _tokenize_mail_addrs {
 
 ###########################################################################
 
+# compute the probability that a token is spammish for each token
+sub _compute_prob_for_all_tokens {
+  my ($self, $tokensdata, $ns, $nn) = @_;
+  my @probabilities;
+
+  return if !$ns || !$nn;
+
+  my $threshold = 1;  # ignore low-freq tokens below this s+n threshold
+  if (!USE_ROBINSON_FX_EQUATION_FOR_LOW_FREQS) {
+    $threshold = 10;
+  }
+  if (!$self->{use_hapaxes}) {
+    $threshold = 2;
+  }
+
+  foreach my $tokendata (@{$tokensdata}) {
+    my $s = $tokendata->[1];  # spam count
+    my $n = $tokendata->[2];  # ham count
+    my $prob;
+
+    no warnings 'uninitialized';  # treat undef as zero in addition
+    if ($s + $n >= $threshold) {
+      # ignoring low-freq tokens, also covers the (!$s && !$n) case
+
+      # my $ratios = $s / $ns;
+      # my $ration = $n / $nn;
+      # $prob = $ratios / ($ration + $ratios);
+      #
+      $prob = ($s * $nn) / ($n * $ns + $s * $nn);  # same thing, faster
+
+      if (USE_ROBINSON_FX_EQUATION_FOR_LOW_FREQS) {
+        # use Robinson's f(x) equation for low-n tokens, instead of just
+        # ignoring them
+        my $robn = $s + $n;
+        $prob =
+          ($Mail::SpamAssassin::Bayes::Combine::FW_S_DOT_X + ($robn * $prob))
+                               /
+          ($Mail::SpamAssassin::Bayes::Combine::FW_S_CONSTANT + $robn);
+      }
+    }
+
+    # 'log_raw_counts' is used to log the raw data for the Bayes equations
+    # during a mass-check, allowing the S and X constants to be optimized
+    # quickly without requiring re-tokenization of the messages for each
+    # attempt. There's really no need for this code to be uncommented in
+    # normal use, however.   It has never been publicly documented, so
+    # commenting it out is fine. ;)
+    #
+    ## if ($self->{log_raw_counts}) {
+    ## $self->{raw_counts} .= " s=$s,n=$n ";
+    ## }
+
+    push(@probabilities, $prob);
+  }
+  return \@probabilities;
+}
+
 # compute the probability that a token is spammish
 sub _compute_prob_for_token {
   my ($self, $token, $ns, $nn, $s, $n) = @_;
@@ -1359,53 +1546,14 @@ sub _compute_prob_for_token {
   # we allow the caller to give us the token information, just
   # to save a potentially expensive lookup
   if (!defined($s) || !defined($n)) {
-    ($s, $n, undef) = $self->{store}->tok_get ($token);
+    ($s, $n, undef) = $self->{store}->tok_get($token);
   }
+  return if !$s && !$n;
 
-  return if ($s == 0 && $n == 0);
+  my $probabilities_ref =
+    $self->_compute_prob_for_all_tokens([ [$token, $s, $n, 0] ], $ns, $nn);
 
-  if (!USE_ROBINSON_FX_EQUATION_FOR_LOW_FREQS) {
-    return if ($s + $n < 10);      # ignore low-freq tokens
-  }
-
-  if (!$self->{use_hapaxes}) {
-    return if ($s + $n < 2);
-  }
-
-  return if ( $ns == 0 || $nn == 0 );
-
-  my $ratios = ($s / $ns);
-  my $ration = ($n / $nn);
-
-  my $prob;
-
-  if ($ratios == 0 && $ration == 0) {
-    warn "bayes: oops? ratios == ration == 0";
-    return;
-  } else {
-    $prob = ($ratios) / ($ration + $ratios);
-  }
-
-  if (USE_ROBINSON_FX_EQUATION_FOR_LOW_FREQS) {
-    # use Robinson's f(x) equation for low-n tokens, instead of just
-    # ignoring them
-    my $robn = $s+$n;
-    $prob = ($Mail::SpamAssassin::Bayes::Combine::FW_S_DOT_X + ($robn * $prob))
-                             /
-            ($Mail::SpamAssassin::Bayes::Combine::FW_S_CONSTANT + $robn);
-  }
-
-  # 'log_raw_counts' is used to log the raw data for the Bayes equations during
-  # a mass-check, allowing the S and X constants to be optimized quickly
-  # without requiring re-tokenization of the messages for each attempt. There's
-  # really no need for this code to be uncommented in normal use, however.   It
-  # has never been publicly documented, so commenting it out is fine. ;)
-
-  ## if ($self->{log_raw_counts}) {
-  ## $self->{raw_counts} .= " s=$s,n=$n ";
-  ## }
-
-  return $prob;
+  return $probabilities_ref->[0];
 }
 
 ###########################################################################
@@ -1497,10 +1645,17 @@ sub learner_new {
   my ($self) = @_;
 
   my $store;
-  my $module = untaint_var($self->{conf}->{bayes_store_module});
-  $module = 'Mail::SpamAssassin::BayesStore::DBM'  if !$module;
+  my $module = $self->{conf}->{bayes_store_module};
+  if (!$module) {
+    $module = 'Mail::SpamAssassin::BayesStore::DBM';
+  } elsif ($module =~ /^([_A-Za-z0-9:]+)$/) {
+    $module = untaint_var($module);
+  } else {
+    die "bayes: invalid module: $module\n";
+  }
 
   dbg("bayes: learner_new self=%s, bayes_store_module=%s", $self,$module);
+  undef $self->{store};  # DESTROYs previous object, if any
   eval '
     require '.$module.';
     $store = '.$module.'->new($self);
