@@ -19,17 +19,17 @@ package Mail::SpamAssassin::Plugin::HeaderEval;
 
 use strict;
 use warnings;
-use bytes;
+# use bytes;
 use re 'taint';
 use Errno qw(EBADF);
 
 use Mail::SpamAssassin::Plugin;
 use Mail::SpamAssassin::Locales;
+use Mail::SpamAssassin::Util qw(get_my_locales parse_rfc822_date);
 use Mail::SpamAssassin::Logger;
 use Mail::SpamAssassin::Constants qw(:sa :ip);
 
-use vars qw(@ISA);
-@ISA = qw(Mail::SpamAssassin::Plugin);
+our @ISA = qw(Mail::SpamAssassin::Plugin);
 
 # constructor: register the eval rule
 sub new {
@@ -52,6 +52,7 @@ sub new {
   $self->register_eval_rule("check_for_forged_eudoramail_received_headers");
   $self->register_eval_rule("check_for_forged_yahoo_received_headers");
   $self->register_eval_rule("check_for_forged_juno_received_headers");
+  $self->register_eval_rule("check_for_forged_gmail_received_headers");
   $self->register_eval_rule("check_for_matching_env_and_hdr_from");
   $self->register_eval_rule("sorted_recipients");
   $self->register_eval_rule("similar_recipients");
@@ -116,7 +117,7 @@ sub check_for_faraway_charset_in_headers {
   my ($self, $pms) = @_;
   my $hdr;
 
-  my @locales = Mail::SpamAssassin::Util::get_my_locales($self->{main}->{conf}->{ok_locales});
+  my @locales = get_my_locales($self->{main}->{conf}->{ok_locales});
 
   return 0 if grep { $_ eq "all" } @locales;
 
@@ -265,17 +266,10 @@ sub word_is_in_dictionary {
 sub check_illegal_chars {
   my ($self, $pms, $header, $ratio, $count) = @_;
 
-  $header .= ":raw" unless ($header eq "ALL" || $header =~ /:raw$/);
+  $header .= ":raw" unless $header =~ /:raw$/;
   my $str = $pms->get($header);
   return 0 if !defined $str || $str eq '';
 
-  # avoid overlap between tests
-  if ($header eq "ALL") {
-    # fix continuation lines, then remove Subject and From
-    $str =~ s/\n[ \t]+/  /gs;
-    $str =~ s/^(?:Subject|From):.*$//gmi;
-  }
-
   # count illegal substrings (RFC 2045)
   # (non-ASCII + C0 controls except TAB, NL, CR)
   my $illegal = $str =~ tr/\x00-\x08\x0b\x0c\x0e-\x1f\x7f-\xff//;
@@ -343,17 +337,24 @@ sub _check_for_forged_hotmail_received_headers {
 
   my $ip = $pms->get('X-Originating-Ip',undef);
   my $IP_ADDRESS = IP_ADDRESS;
+  my $orig = $pms->get('X-OriginatorOrg',undef);
+  my $ORIGINATOR = 'hotmail.com';
 
   if (defined $ip && $ip =~ /$IP_ADDRESS/) { $ip = 1; } else { $ip = 0; }
+  if (defined $orig && $orig =~ /$ORIGINATOR/) { $orig = 1; } else { $orig = 0; }
 
   # Hotmail formats its received headers like this:
   # Received: from hotmail.com (f135.law8.hotmail.com [216.33.241.135])
+  # or like
+  # Received: from EUR01-VE1-obe.outbound.protection.outlook.com (mail-oln040092066056.outbound.protection.outlook.com [40.92.66.56])
   # spammers do not ;)
 
   if ($self->gated_through_received_hdr_remover($pms)) { return; }
 
   if ($rcvd =~ /from (?:\S*\.)?hotmail.com \(\S+\.hotmail(?:\.msn)?\.com[ \)]/ && $ip)
                 { return; }
+  if ($rcvd =~ /from \S*\.outbound\.protection\.outlook\.com \(\S+\.outbound\.protection\.outlook\.com[ \)]/ && $orig)
+                { return; }
   if ($rcvd =~ /from \S*\.hotmail.com \(\[$IP_ADDRESS\][ \):]/ && $ip)
                 { return; }
   if ($rcvd =~ /from \S+ by \S+\.hotmail(?:\.msn)?\.com with HTTP\;/ && $ip)
@@ -515,6 +516,7 @@ sub check_for_forged_yahoo_received_headers {
             { return 0; }
 
   if ($rcvd =~ /by web\S+\.mail\S*\.yahoo\.com via HTTP/) { return 0; }
+  if ($rcvd =~ /by sonic\S+\.consmr\.mail\S*\.yahoo\.com with HTTP/) { return 0; }
   if ($rcvd =~ /by smtp\S+\.yahoo\.com with SMTP/) { return 0; }
   my $IP_ADDRESS = IP_ADDRESS;
   if ($rcvd =~
@@ -573,6 +575,32 @@ sub check_for_forged_juno_received_headers {
   return 0;   
 }
 
+sub check_for_forged_gmail_received_headers {
+  my ($self, $pms) = @_;
+  use constant GOOGLE_MESSAGE_STATE_LENGTH_MIN => 60;
+  use constant GOOGLE_SMTP_SOURCE_LENGTH_MIN => 60;
+
+  my $from = $pms->get('From:addr');
+  if ($from !~ /\bgmail\.com$/i) { return 0; }
+
+  my $xgms = $pms->get('X-Gm-Message-State');
+  my $xss = $pms->get('X-Google-Smtp-Source');
+  my $xreceived = $pms->get('X-Received');
+  my $received = $pms->get('Received');
+
+  if ($xreceived =~ /by 10\.\S+ with SMTP id \S+/) { return 0; }
+  if ($xreceived =~ /by 2002\:a\d\d\:\w+\:\S+ with SMTP id \S+/) { return 0; }
+  if ($received =~ /by smtp\.googlemail\.com with ESMTPSA id \S+/) {
+    return 0;
+  }
+  if ( (length($xgms) >= GOOGLE_MESSAGE_STATE_LENGTH_MIN) && 
+    (length($xss) >= GOOGLE_SMTP_SOURCE_LENGTH_MIN)) {
+      return 0;
+  }
+
+  return 1;
+}
+
 sub check_for_matching_env_and_hdr_from {
   my ($self, $pms) =@_;
   # two blank headers match so don't bother checking
@@ -722,7 +750,7 @@ sub _get_date_header_time {
     for my $date (@dates) {
       if (defined($date) && length($date)) {
         chomp($date);
-        $time = Mail::SpamAssassin::Util::parse_rfc822_date($date);
+        $time = parse_rfc822_date($date);
       }
       last DATE if defined($time);
     }
@@ -773,7 +801,7 @@ sub _get_received_header_times {
       if ($rcvd =~ m/(\s.?\d+ \S\S\S \d+ \d+:\d+:\d+ \S+)/) {
 	my $date = $1;
         dbg2("eval: trying Received fetchmail header date for real time: $date");
-	my $time = Mail::SpamAssassin::Util::parse_rfc822_date($date);
+	my $time = parse_rfc822_date($date);
 	if (defined($time) && (time() >= $time)) {
           dbg2("eval: time_t from date=$time, rcvd=$date");
 	  push @fetchmail_times, $time;
@@ -793,7 +821,7 @@ sub _get_received_header_times {
     if ($rcvd =~ m/(\s.?\d+ \S\S\S \d+ \d+:\d+:\d+ \S+)/) {
       my $date = $1;
       dbg2("eval: trying Received header date for real time: $date");
-      my $time = Mail::SpamAssassin::Util::parse_rfc822_date($date);
+      my $time = parse_rfc822_date($date);
       if (defined($time)) {
         dbg2("eval: time_t from date=$time, rcvd=$date");
 	push @header_times, $time;
@@ -894,7 +922,7 @@ sub subject_is_all_caps {
 
    $subject =~ s/^\s+//;
    $subject =~ s/\s+$//;
-   $subject =~ s/^(?:(?:Re|Fwd|Fw|Aw|Antwort|Sv):\s*)+//i;  # Bug 6805
+   $subject =~ s/^(?:(?:Re|Fwd|Fw|Aw|Antwort|Sv|VS):\s*)+//i;  # Bug 6805
    return 0 if $subject !~ /\s/;	# don't match one word subjects
    return 0 if (length $subject < 10);  # don't match short subjects
    $subject =~ s/[^a-zA-Z]//g;		# only look at letters
@@ -953,14 +981,14 @@ sub check_outlook_message_id {
   my $fudge = 250;
 
   $_ = $pms->get('Date');
-  $_ = Mail::SpamAssassin::Util::parse_rfc822_date($_) || 0;
+  $_ = parse_rfc822_date($_) || 0;
   my $expected = int (($_ * $x) + $y);
   my $diff = $timetoken - $expected;
   return 0 if (abs($diff) < $fudge);
 
   $_ = $pms->get('Received');
   /(\s.?\d+ \S\S\S \d+ \d+:\d+:\d+ \S+).*?$/;
-  $_ = Mail::SpamAssassin::Util::parse_rfc822_date($_) || 0;
+  $_ = parse_rfc822_date($_) || 0;
   $expected = int(($_ * $x) + $y);
   $diff = $timetoken - $expected;
 
@@ -1000,7 +1028,6 @@ sub check_unresolved_template {
   my ($self, $pms) = @_;
 
   my $all = $pms->get('ALL');	# cached access
-  $all =~ s/\n[ \t]+/ /gs;	# fix continuation lines
   
   for my $header (split(/\n/, $all)) {
     # slightly faster to test in this order
@@ -1033,9 +1060,9 @@ sub check_ratware_envelope_from {
   my $from = $pms->get('EnvelopeFrom:addr');
 
   return 0 if $from eq '' || $to eq '';
-  return 0 if $from =~ /^SRS\d=/;
+  return 0 if $from =~ /^SRS\d[=+-]/i;
 
-  if ($to =~ /^([^@]+)@(.+)$/) {
+  if ($to =~ /^([^@]+)\@(.+)$/) {
     my($user,$dom) = ($1,$2);
     $dom = $self->{main}->{registryboundaries}->trim_domain($dom);
     return unless
